"""
Atomica version file.

Standard location for module version number and date.
"""

import sciris as sc

version = "1.11.0"
versiondate = "2019-03-07"
gitinfo = sc.gitinfo(__file__, verbose=False)

